﻿using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace StackMachine.Ast
{
    /// <summary>
    /// 関数定義。
    /// </summary>
    public class Function : Statement
    {
        /// <summary>
        /// 関数名。
        /// </summary>
        public string Name { get; internal set; }

        /// <summary>
        /// 仮引数リスト。
        /// </summary>
        public List<Parameter> Arguments { get; internal set; }

        /// <summary>
        /// 関数本体。
        /// </summary>
        public Block Body { get; internal set; }

        /// <summary>
        /// 関数内で定義されたローカル変数一覧。
        /// </summary>
        public IEnumerable<Parameter> LocalVariables
        {
            get
            {
                return this.Body.LocalVariables;
            }
        }

        /// <summary>
        /// 関数＋引数名、
        /// let f(x, y) = x + y; の f(x, y) の部分を文字列として返す。
        /// </summary>
        /// <returns>関数＋引数名</returns>
        public string Signature()
        {
            StringBuilder sb = new StringBuilder();

            sb.Append(this.Name);
            sb.Append("(");

            bool first = true;

            foreach (var arg in this.Arguments)
            {
                if (first) first = false;
                else sb.Append(", ");

                sb.Append(arg.Name);
            }

            sb.Append(")");

            return sb.ToString();
        }

        public override NodeType Type
        {
            get { return NodeType.Function; }
        }

        public override bool Equals(object obj)
        {
            var x = obj as Function;

            if (x == null) return false;

            if (this.Name != x.Name) return false;

            if (this.Arguments.Count != x.Arguments.Count) return false;

            for (int i = 0; i < this.Arguments.Count; i++)
            {
                if (!this.Arguments[i].Equals(x.Arguments[i])) return false;
            }

            return this.Body.Equals(x.Body);
        }

        public override int GetHashCode()
        {
            return this.Name.GetHashCode() ^ this.Arguments.GetHashCode() ^ this.Body.GetHashCode();
        }

        protected internal override IEnumerable<StackMachine.Emurator.Instruction> Compile(Dictionary<Parameter, int> localTable, Dictionary<Function, int> addressTable, Dictionary<Function, IEnumerable<StackMachine.Emurator.Instruction>> functionTable)
        {
            localTable = new Dictionary<Parameter, int>();

            int i = 0;
            foreach (var arg in this.Arguments)
            {
                localTable[arg] = i;
                ++i;
            }

            functionTable[this] = this.Body.Compile(localTable, addressTable, functionTable).ToList();

            return Enumerable.Range(0, 0).Select(x => StackMachine.Emurator.Instruction.Nop);
        }
    }
}
